#include <X11/Wc/COPY>

#include <Xm/Xm.h>		/* Motif and Xt Intrinsics	*/
#include <X11/Wc/WcCreate.h>	/* Widget Creation Library	*/
#include <X11/Xmp/Xmp.h>		/* Motif Public widgets etc.	*/

#include <BossP.h>

/******************************************************************************
**  Private Data
******************************************************************************/

/* All Wcl applications should provide at least these Wcl options:
*/
static XrmOptionDescRec options[] = {
    WCL_XRM_OPTIONS
};

/******************************************************************************
**  Private Functions
******************************************************************************/

/*ARGSUSED*/
static void DeleteWindowCB( w, clientData, callData )
    Widget	w;
    XtPointer	clientData;
    XtPointer	callData;
{
    /* This callback is invoked when the user selects `Close' from
    ** the mwm frame menu on the upper left of the window border.
    ** Do whatever is appropriate.
    */
    printf("Closed by window manager.\n");
    exit(0);
}

/*
 * layout buttons equally space and of the largest since of any one
 */
void DoButtonLayout( XmBossWidget w )
{
    Dimension maxWidth = 1, maxHeight = 1;
    int i, numManaged=0;
    Widget child;
    int extraWidth, extraHeight;
    int x;
    XtWidgetGeometry desired;

    for (i=0; i<w->composite.num_children; i++)
    {
	child = w->composite.children[i];
	if ( XtIsManaged(child) )
	{
	    numManaged++;
	    XtQueryGeometry( child, NULL, &desired );
printf("Child '%s'  desires w=%d h=%d\n", XtName(child), desired.width, desired.height);
	    if ( desired.width > maxWidth )
		maxWidth = desired.width;
	    if ( desired.height > maxHeight )
		maxHeight = desired.height;
	}
    }
    if ( numManaged )
    {
	/* no size set yet? ask for our ideal size */
	if ( w->core.width <= 1 && w->core.height <= 1 )
	{
	    XtMakeResizeRequest( w, 
			10 + ((10 + maxWidth) * numManaged),
			20 + maxHeight, 
			NULL,
			NULL );
	}
	extraWidth  = ( w->core.width - (maxWidth * numManaged) );
	extraHeight = ( w->core.height - maxHeight );
	if ( extraWidth < 0 )
	{
	    extraWidth = 0;
	    maxWidth = w->core.width / numManaged;
	}
	else
	extraWidth /= ( numManaged + 1 );
	if ( extraHeight < 0 )
	{
	    extraHeight = 0;
	    maxHeight = w->core.height;
	}
	extraHeight /= 2;

	x = extraWidth;
	for (i=0; i<w->composite.num_children; i++)
	{
	    child = w->composite.children[i];
	    if ( XtIsManaged(child) )
	    {
		_XmConfigureObject( (RectObj)child, x, extraHeight, 
		    maxWidth, maxHeight, 0 );
		x += (extraWidth + maxWidth);
	    }
	}
    }
}

/*
 * layout widget pairs (label/TextField) with the left widget being
 * right justified and the right widget being left justified.  The
 * pairs are arranged in a column formation.
 */
typedef struct 
{
    Dimension	w, h;
} LayoutBox;

void DoEntryLayout( XmBossWidget wid )
{
    int i;
    Widget child;
    XtWidgetGeometry desired;
    int x, y, w, h;

    int totalHeight=0;
    int maxHeight=0, maxLeftWidth=0, maxRightWidth=0;
    int column=0;
    int rows=0;
    int numManaged = 0;
    int extraWidth, extraHeight;
    int limitRightWidth = 0;

    LayoutBox *layouts;
    int currLayout=0;

    if ( wid->composite.num_children <= 0 )
	return;

    totalHeight += 2*MGR_ShadowThickness(wid);

    layouts = (LayoutBox *)XtMalloc( wid->composite.num_children * sizeof(LayoutBox) );

    for (i=0; i<wid->composite.num_children; i++)
    {
	child = wid->composite.children[i];
	if ( XtIsManaged(child) )
	{
	    if ( column == 0)
	    {
		maxHeight = 1;
	    }
	    numManaged++;
	    XtQueryGeometry( child, NULL, &desired );
printf("Child '%s'  desires w=%d h=%d\n", XtName(child), desired.width, desired.height);
	    if ( desired.height > maxHeight )
		maxHeight = desired.height;
	    if ( column ==  0)
	    {
		if ( desired.width > maxLeftWidth )
		    maxLeftWidth = desired.width;
		column = 1;
		layouts[currLayout++].w = desired.width;
		rows++;
	    }
	    else
	    {
		if ( desired.width > maxRightWidth )
		    maxRightWidth = desired.width;
		totalHeight += maxHeight;

		layouts[currLayout-1].h = maxHeight;
		layouts[currLayout].h = maxHeight;
		layouts[currLayout++].w = desired.width;
		/* reset for next row */
		maxHeight = 1;
		column = 0;
	    }
	}
    }
    /* end up with odd # entries? */
    if ( column == 1 )
    {
printf("odd column: currLayout=%d maxHeight = %d\n", currLayout, maxHeight );
	totalHeight += maxHeight;
	layouts[currLayout-1].h = maxHeight;
    }

    if ( numManaged )
    {
	/* no size set yet? ask for our ideal size */
	if ( wid->core.width <= 1 && wid->core.height <= 1 )
	{
printf("totalHeight=%d  maxLeftWidth=%d  maxRightWidth=%d\n", totalHeight, maxLeftWidth, maxRightWidth);
	    XtMakeResizeRequest( wid, 
		maxLeftWidth + maxRightWidth + (2*MGR_ShadowThickness(wid)),
		totalHeight,
		NULL,
		NULL );
	}
	extraWidth = wid->core.width - 
		(maxLeftWidth + maxRightWidth + (2*MGR_ShadowThickness(wid)));
	extraHeight = wid->core.height - totalHeight;
	if ( extraWidth > 0 )
	{
	    extraWidth /= 3;
	    limitRightWidth = 9999;
	}
	else
	{
	    limitRightWidth = maxRightWidth + extraWidth;
	    if ( limitRightWidth <= 0 )
	    {
		limitRightWidth = 9999;
	    }
	    extraWidth = 0;
	}

	if ( extraHeight > 0 )
	    extraHeight /= rows+1;
	else
	    extraHeight = 0;

	currLayout = 0;
	column = 0;
	y = MGR_ShadowThickness(wid) + extraHeight;;
	for (i=0; i<wid->composite.num_children; i++)
	{
	    child = wid->composite.children[i];

	    if ( XtIsManaged(child) )
	    {
printf("layout[%d] : w=%d h=%d\n", currLayout, layouts[currLayout].w, layouts[currLayout].h);
		w = layouts[currLayout].w;
		if ( w > limitRightWidth )
		    w = limitRightWidth;

		h = layouts[currLayout++].h;
		if ( column == 0 )
		    x = maxLeftWidth - w + (extraWidth / 2);
		else
		    x = maxLeftWidth + (extraWidth / 2);
		x += MGR_ShadowThickness(wid) + extraWidth;

		_XmConfigureObject( (RectObj)child, x, y, w, h, 0 );
		if ( column == 1 )
		{
		    y += h + extraHeight;
		    column = 0;
		}
		else
		    column = 1;
	    }
	}
    }
    XtFree( layouts );
}

static void DoChangeManaged( w, clientData, callData )
    Widget	w;
    XtPointer	clientData;
    XtPointer	callData;
{
    /* DoButtonLayout( (XmBossWidget)w ); */
    DoEntryLayout( (XmBossWidget)w );
}

static void DoResize( w, clientData, callData )
    Widget	w;
    XtPointer	clientData;
    XtPointer	callData;
{
    /* DoButtonLayout( (XmBossWidget)w ); */
    DoEntryLayout( (XmBossWidget)w );
}

static void DoGeometryManager( w, clientData, cb )
    Widget	w;
    XtPointer	clientData;
    XmBossGeometryManagerCallbackStruct	*cb;
{
    cb->result = XtGeometryNo;
}

/*ARGSUSED*/
static void RegisterApplication ( app )
    XtAppContext app;
{
    /* -- Useful shorthand for registering things with the Wcl library */
#define RCP( name, class  ) WcRegisterClassPtr   ( app, name, class );
#define RCO( name, constr ) WcRegisterConstructor( app, name, constr );
#define RAC( name, func   ) WcRegisterAction     ( app, name, func );
#define RCB( name, func   ) WcRegisterCallback   ( app, name, func, NULL );

    /* -- register widget classes and constructors */
    RCO( "XmCreateBoss", XmCreateBoss );
    RCB( "DoGeometryManager", DoGeometryManager);
    RCB( "DoResize", DoResize);
    RCB( "DoChangeManaged", DoChangeManaged);

    /* -- Register application specific actions */
    /* -- Register application specific callbacks */
}


/******************************************************************************
*   MAIN function
******************************************************************************/

main ( argc, argv )
    int     argc;
    String  argv[];
{   

    /*  -- Intialize Toolkit creating the application shell
    */
    Widget appShell = XtInitialize ( 
	argv[0], WcAppClass( argc, argv ),	/* application name & class */
	options, XtNumber(options),		/* descr of cmd line options */
	&argc, argv 
    );
    XtAppContext app = XtWidgetToApplicationContext(appShell);

    /*  -- Register all application specific callbacks and widget classes
    */
    RegisterApplication ( app );

    /*  -- Register all Motif classes, constructors, and Xmp CBs & ACTs
    */
    XmpRegisterAll ( app );

    /*  -- Create widget tree below toplevel shell using Xrm database
    */
    if ( WcWidgetCreation ( appShell ) )
	exit(1);

    /*  -- Realize the widget tree
    */
    XtRealizeWidget ( appShell );

    /*  -- Optional, but frequently desired:  
    ** Provide a callback which gets invoked when the user selects `Close' from
    ** the mwm frame menu on the top level shell.  A real application will need
    ** to provide its own callback instead of DeleteWindowCB, and probably 
    ** client data too.  MUST be done after shell widget is REALIZED!  Hence,
    ** this CANNOT be done using wcCallback (in a creation time callback).
    */
    XmpAddMwmCloseCallback( appShell, DeleteWindowCB, NULL );

    /*  -- and finally, enter the main application loop
    */
    XtMainLoop ( );
}
